## Copyright 2017 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import sys
import os
import logging
import subprocess
import time
import json
import traceback
import six
from six.moves.urllib import parse as urlparse

logging.basicConfig(level=logging.DEBUG, format='%(levelname)s:%(threadName)s:%(module)s.%(funcName)s: %(message)s')

# We have to be in the correct directory *before* we import clibs so we're going to do this as early as possible.
if hasattr(sys, 'frozen'):
    if hasattr(sys, '_MEIPASS'):
        os.chdir(sys._MEIPASS)
    else:
        os.chdir(os.path.dirname(sys.executable))
else:
    my_path = os.path.dirname(__file__)
    if my_path != '':
        os.chdir(my_path)


from . import center

# Initialize the FileHandler early to capture all log messages.
if not os.path.isdir(center.settings_path):
    os.makedirs(center.settings_path)

# We truncate the log file on every start to avoid filling the user's disk with useless data.
log_path = os.path.join(center.settings_path, 'log.txt')

try:
    if os.path.isfile(log_path):
        os.unlink(log_path)
except Exception:
    # This will only be visible if the user is running a console version.
    logging.exception('The log is in use by someone!')
else:
    handler = logging.FileHandler(log_path, 'w')
    handler.setFormatter(logging.Formatter('%(levelname)s:%(threadName)s:%(module)s.%(funcName)s: %(message)s'))
    handler.setLevel(logging.DEBUG)
    logging.getLogger().addHandler(handler)

if not center.DEBUG:
    logging.getLogger().setLevel(logging.INFO)

if six.PY2:
    from . import py2_compat  # noqa

from .qt import QtCore, QtGui, QtWidgets, variant as qt_variant
from .tasks import run_task, CheckUpdateTask
from . import util, ipc, auto_fetch


app = None
ipc_conn = None
translate = QtCore.QCoreApplication.translate


def my_excepthook(type, value, tb):
    try:
        # NOTE: This can fail (for some reason) in traceback.print_exception.
        logging.error('UNCAUGHT EXCEPTION!', exc_info=(type, value, tb))
    except Exception:
        logging.error('UNCAUGHT EXCEPTION!\n%s%s: %s' % (''.join(traceback.format_tb(tb)), type.__name__, value))

    msg = translate('launcher', 'A critical error occurred! Knossos might not work correctly until you restart it.')
    if center.raven:
        msg += '\n' + translate('launcher', 'The error has been reported and will hopefully be fixed soon.')

    try:
        QtWidgets.QMessageBox.critical(None, 'Knossos', msg)
    except Exception:
        pass


def get_cmd(args=[]):
    if hasattr(sys, 'frozen'):
        my_path = [os.path.abspath(sys.executable)]
    else:
        my_path = [os.path.abspath(sys.executable), os.path.abspath('__main__.py')]

    return my_path + args


def get_file_path(name):
    if hasattr(sys, 'frozen') or os.path.isdir('data'):
        return os.path.join('data', name)
    else:
        from pkg_resources import resource_filename
        return resource_filename(__package__, name)


def load_settings():
    spath = os.path.join(center.settings_path, 'settings.json')
    settings = center.settings
    if os.path.exists(spath):
        defaults = settings.copy()

        try:
            with open(spath, 'r') as stream:
                settings.update(json.load(stream))
        except Exception:
            logging.exception('Failed to load settings from "%s"!', spath)

        # Migration
        if 's_version' not in settings:
            settings['s_version'] = 0

        if settings['s_version'] < 3:
            if 'mods' in settings:
                del settings['mods']
            if 'installed_mods' in settings:
                del settings['installed_mods']

            settings['s_version'] = 3

        if settings['s_version'] < 5:
            settings['repos'] = defaults['repos']
            settings['nebula_link'] = defaults['nebula_link']
            settings['s_version'] = 5

        del defaults
    else:
        # Most recent settings version
        settings['s_version'] = 5

    if settings['hash_cache'] is not None:
        util.HASH_CACHE = settings['hash_cache']

    if settings['use_raven']:
        util.enable_raven()

    return settings


def run_knossos():
    global app

    from . import repo, progress, integration
    from .windows import HellWindow

    if sys.platform.startswith('win') and os.path.isfile('7z.exe'):
        util.SEVEN_PATH = os.path.abspath('7z.exe')
    elif sys.platform == 'darwin' and os.path.isfile('7z'):
        util.SEVEN_PATH = os.path.abspath('7z')

    translate = QtCore.QCoreApplication.translate

    if not util.test_7z():
        QtWidgets.QMessageBox.critical(None, 'Knossos', translate(
            'launcher', 'I can\'t find "7z"! Please install it and run this program again.'))
        return

    util.DL_POOL.set_capacity(center.settings['max_downloads'])
    if center.settings['download_bandwidth'] > 0.0:
        util.SPEED_LIMIT_BUCKET.set_rate(center.settings['download_bandwidth'])

    center.app = app
    center.installed = repo.InstalledRepo()
    center.pmaster = progress.Master()
    center.pmaster.start_workers(10)
    center.mods = repo.Repo()
    center.auto_fetcher = auto_fetch.AutoFetcher()

    # This has to run before we can load any mods!
    repo.CPU_INFO = util.get_cpuinfo()

    integration.init()
    mod_db = os.path.join(center.settings_path, 'mods.json')
    if os.path.isfile(mod_db):
        center.mods.load_json(mod_db)

    center.main_win = HellWindow()
    center.main_win.open()

    if center.settings['update_notify'] and '-dev' not in center.VERSION:
        run_task(CheckUpdateTask())

    app.exec_()

    center.save_settings()
    ipc.shutdown()


def handle_ipc_error():
    global app, ipc_conn

    logging.warning('Failed to connect to main process!')

    if ipc_conn is not None:
        ipc_conn.clean()
        ipc_conn = None


def scheme_handler(link):
    global app, ipc_conn

    if not link.startswith(('fs2://', 'fso://')):
        # NOTE: fs2:// is deprecated, we don't tell anyone about it.
        QtWidgets.QMessageBox.critical(None, 'Knossos',
            translate('launcher', 'I don\'t know how to handle "%s"! I only know fso:// .') % (link))
        app.quit()
        return True

    link = urlparse.unquote(link.strip()).split('/')

    if len(link) < 3:
        QtWidgets.QMessageBox.critical(None, 'Knossos', translate('launcher', 'Not enough arguments!'))
        app.quit()
        return True

    if not ipc_conn.server_exists():
        # Launch the program.
        subprocess.Popen(get_cmd())

        # Wait for the program...
        start = time.time()
        while not ipc_conn.server_exists():
            if time.time() - start > 20:
                # That's too long!
                QtWidgets.QMessageBox.critical(None, 'Knossos', translate('launcher', 'Failed to start server!'))
                app.quit()
                return True

            time.sleep(0.3)

    try:
        ipc_conn.open_connection(handle_ipc_error)
    except Exception:
        logging.exception('Failed to connect to myself!')
        handle_ipc_error()
        return False

    if not ipc_conn:
        return False

    ipc_conn.send_message(json.dumps(link[2:]))
    ipc_conn.close(True)
    app.quit()
    return True


def main():
    global ipc_conn, app

    # Default to replacing errors when de/encoding.
    import codecs

    codecs.register_error('strict', codecs.replace_errors)
    codecs.register_error('really_strict', codecs.strict_errors)

    sys.excepthook = my_excepthook

    # The version file is only read in dev builds.
    if center.VERSION.endswith('-dev'):
        if 'KN_VERSION' in os.environ:
            center.VERSION = os.environ['KN_VERSION'].strip()
        else:
            try:
                with open('../.git/HEAD') as stream:
                    ref = stream.read().strip().split(':')
                    assert ref[0] == 'ref'

                with open('../.git/' + ref[1].strip()) as stream:
                    center.VERSION += '+' + stream.read()[:7]
            except Exception:
                pass

    logging.info('Running Knossos %s on %s and Python %s.', center.VERSION, qt_variant, sys.version)
    app = QtWidgets.QApplication([])

    res_path = get_file_path('resources.rcc')
    logging.debug('Loading resources from %s.', res_path)
    QtCore.QResource.registerResource(res_path)

    logging.debug('Loading settings...')
    load_settings()
    
    trans = QtCore.QTranslator()
    if center.settings['language']:
        lang = center.settings['language']
    else:
        lang = QtCore.QLocale()

    if trans.load(lang, 'knossos', '_', get_file_path(''), '.etak'):
        app.installTranslator(trans)
    else:
        del trans

    app.setWindowIcon(QtGui.QIcon(':/hlp.png'))
    ipc_conn = ipc.IPCComm(center.settings_path)

    if len(sys.argv) > 1:
        if sys.argv[1] == '--install-scheme':
            from . import integration

            integration.init()
            integration.install_scheme_handler('--silent' not in sys.argv)
            return
        elif sys.argv[1] == '--finish-update':
            updater = sys.argv[2]

            if not os.path.isfile(updater):
                logging.error('The update finished but where is the installer?! It\'s not where it\'s supposed to be! (%s)', updater)
            else:
                tries = 3
                while tries > 0:
                    try:
                        # Clean up
                        os.unlink(updater)
                    except Exception:
                        logging.exception('Failed to remove updater! (%s)' % updater)

                    if os.path.isfile(updater):
                        time.sleep(0.3)
                        tries -= 1
                    else:
                        break

                # Delete the temporary directory.
                if os.path.basename(updater) == 'knossos_updater.exe':
                    try:
                        os.rmdir(os.path.dirname(updater))
                    except Exception:
                        logging.exception('Failed to remove the updater\'s temporary directory.')
        else:
            tries = 3
            while tries > 0:
                if scheme_handler(sys.argv[1]):
                    break

                tries -= 1
                ipc_conn = ipc.IPCComm(center.settings_path)

            if tries == 0:
                sys.exit(1)

            return
    elif ipc_conn.server_exists() and scheme_handler('fso://focus'):
        return

    del ipc_conn

    try:
        run_knossos()
    except Exception:
        logging.exception('Uncaught exeception! Quitting...')

        # Try to tell the user
        QtWidgets.QMessageBox.critical(None, 'Knossos',
            translate('launcher', 'I encountered a fatal error.\nI\'m sorry but I\'m going to crash now...'))



## Copyright 2017 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import sys
import logging
import threading
import six

from . import uhf
uhf(__name__)

from .qt import QtCore

try:
    import curses
except ImportError:
    curses = None

if six.PY2:
    threading.get_ident = lambda: threading.current_thread().ident

_progress = threading.local()


def reset():
    global _progress

    _progress.value = 0.0
    _progress.text = ''
    _progress.tasks = []
    _progress.callback = None


# Initialize with empty values.
reset()


# total_progress = off + span * task_progress
# Example: If you're tracking downloads and starting file 3 of 10, call start_task(3/10, 'Downloading whatever (%s)...')
def start_task(off, span, tmpl='%s'):
    if not hasattr(_progress, 'tasks'):
        reset()

    _progress.tasks.insert(0, (off, float(span), tmpl))


def finish_task():
    if not hasattr(_progress, 'tasks'):
        reset()

    _progress.tasks.pop(0)


def set_callback(cb):
    if not hasattr(_progress, 'tasks'):
        reset()

    _progress.callback = cb


def update(prog, text=''):
    global _progress

    if not hasattr(_progress, 'tasks'):
        reset()

    for task in _progress.tasks:
        prog = task[0] + prog * task[1]
        if '%s' in task[2]:
            text = task[2] % (text,)
        else:
            text = task[2]

    _progress.value = prog
    _progress.text = text

    if _progress.callback is not None:
        _progress.callback(prog, text)


# Task scheduler
class Worker(threading.Thread):
    busy = False

    def __init__(self, master):
        super(Worker, self).__init__()

        self._master = master
        self.daemon = True
        self.start()

    def run(self):
        while True:
            task = self._master._get_work()
            if task is None:
                return

            self.busy = True
            try:
                reset()
                set_callback(task[0]._track_progress)
                task[0]._init()
                task[0].work(*task[1])
            except SystemExit:
                self.busy = False
                return
            except Exception:
                logging.exception('Exception in Thread!')

            task[0]._deinit()
            self.busy = False


class Master(object):
    _tasks = None
    _tasks_lock = None
    _workers = None
    _stop_workers = False
    _worker_cond = None

    def __init__(self):
        self._tasks = []
        self._tasks_lock = threading.Lock()
        self._workers = []
        self._worker_cond = threading.Condition()

    def start_workers(self, num):
        for n in range(0, num):
            self._workers.append(Worker(self))

    def stop_workers(self):
        self._stop_workers = True

        with self._worker_cond:
            self._worker_cond.notify_all()

        for w in self._workers:
            w.join()

        self._stop_workers = False
        self._workers = []
        self._tasks = []

    def _get_work(self):
        while True:
            if self._stop_workers:
                return None

            with self._tasks_lock:
                for task in self._tasks:
                    work = task._get_work()
                    if work is not None:
                        return work

            # No work here... let's wait for more.
            with self._worker_cond:
                self._worker_cond.wait()

    def add_task(self, task):
        if not task._has_work():
            logging.warning('Added an empty task of type "%s". Ignoring it!', task.__class__.__name__)

            # Make sure it finishes.
            task._done.set()
            task.done.emit()
            return

        with self._tasks_lock:
            self._tasks.append(task)
            task._master = self
            task._attached = True

        task.done.connect(self.check_tasks)

        with self._worker_cond:
            self._worker_cond.notify_all()

    def check_tasks(self):
        with self._tasks_lock:
            for task in self._tasks[:]:
                if not task._has_work():
                    self._tasks.remove(task)
                    task._attached = False

    def wake_workers(self):
        with self._worker_cond:
            self._worker_cond.notify_all()

    def is_busy(self):
        return any([w.busy for w in self._workers])


class Task(QtCore.QObject):
    _results = None
    _result_lock = None
    _work = None
    _work_count = 0
    _work_lock = None
    _done = None
    _master = None
    _attached = False
    _progress = None
    _progress_lock = None
    _running = 0
    _pending = 0
    _threads = 0
    _local = None
    _slot_prog = None
    background = False
    can_abort = True
    aborted = False
    title = None
    mods = None
    done = QtCore.Signal()
    progress = QtCore.Signal(tuple)

    def __init__(self, work=None, threads=0):
        super(Task, self).__init__()

        if work is None:
            work = []

        self._results = []
        self._work = work
        self._work_count = len(work)
        self._result_lock = threading.Lock()
        self._work_lock = threading.Lock()
        self._done = threading.Event()
        self._progress_lock = threading.Lock()
        self._threads = threads
        self._local = threading.local()
        self._slot_prog = {}
        self.mods = []

    def _get_work(self):
        with self._work_lock:
            if len(self._work) == 0:
                return None
            elif self._threads > 0 and self._running >= self._threads:
                return None
            else:
                self._pending += 1
                return (self, (self._work.pop(0),))

    def _has_work(self):
        with self._work_lock:
            return len(self._work) > 0

    def _init(self):
        with self._progress_lock:
            self._running += 1

    def _deinit(self):
        with self._progress_lock:
            with self._work_lock:
                self._pending -= 1

            self._running -= 1

            if self._running == 0 and not self._has_work() and not self._done.is_set():
                self._done.set()
                self.done.emit()

    def _track_progress(self, prog, text):
        with self._progress_lock:
            if hasattr(self._local, 'slot'):
                self._slot_prog[self._local.slot] = (self._slot_prog[self._local.slot][0], prog, text)

            total = 0
            for label, prog, text in self._slot_prog.values():
                total += prog

            self.progress.emit((total / max(1, len(self._slot_prog)), self._slot_prog, self.title))

    def post(self, result):
        with self._result_lock:
            self._results.append(result)

    def add_work(self, work):
        with self._work_lock:
            self._work.extend(work)
            self._work_count = max(self._work_count, len(self._work))

        if self._master is not None:
            if not self._attached:
                self._master.add_task(self)
            else:
                self._master.wake_workers()

    def abort(self):
        if not self.can_abort:
            logging.debug("Abort request failed for task %s because it can't abort!" % self)
            return False

        # Empty the work queue, this won't stop running workers but it will
        # stop calls to the work() method.
        with self._work_lock:
            self._work = []
            self.aborted = True

        self._master.check_tasks()

    def get_progress(self):
        with self._work_lock:
            wc_left = len(self._work)
            wc_total = self._work_count
            pending = self._pending

        count = float(wc_total)
        if count == 0:
            count = 0.00001
            total = 1
        else:
            total = (wc_total - wc_left - pending) / count

        for item in prog.values():
            total += item[0] * (1.0 / count)

        return total, {}, self.title

    def is_done(self):
        if not self._done.is_set():
            with self._progress_lock:
                if self._running == 0 and not self._has_work():
                    self._done.set()

        return self._done.is_set()

    def get_results(self):
        if not self._done.is_set():
            self._done.wait()

        with self._result_lock:
            return self._results


class MultistepTask(Task):
    _steps = None
    _sdone = False
    _cur_step = -1

    def __init__(self, steps=None, **kwargs):
        super(MultistepTask, self).__init__(**kwargs)

        if steps is None:
            steps = self._steps

        if isinstance(steps, int):
            snum = steps
            steps = []
            for i in range(1, snum + 1):
                steps.append((getattr(self, 'init' + str(i)), getattr(self, 'work' + str(i))))

        self._steps = steps

    def _has_work(self):
        return not self._sdone and not self.aborted

    def _get_work(self):
        with self._work_lock:
            if (self._threads > 0 and self._pending >= self._threads) or self._sdone or self.aborted:
                return None
            elif len(self._work) == 0 or self._cur_step < 0:
                if self._pending == 0 and self._running == 0:
                    self._pending += 1
                    return (self, ('MAGIC_MULTITASK_STEP_KEY_###',))
                else:
                    return None
            else:
                self._pending += 1
                return (self, (self._work.pop(0),))

    def work(self, arg):
        # Any better ideas for this magic key?
        if arg == 'MAGIC_MULTITASK_STEP_KEY_###':
            # Maybe we need to advance to the next step.
            self._work_lock.acquire()

            if (self._pending == 1 and self._running == 1 and len(self._work) == 0) or self._cur_step < 0:
                self._work_lock.release()
                self._next_step()
            else:
                # TODO: This still happens on Windows and Mac OS. For some reason it doesn't happen on Linux...
                logging.warning('Either we still have some work to do (unlikely) or there are still some other threads running (%d).', self._running)
                self._work_lock.release()

            return

        # Call the current work method
        self._steps[self._cur_step][1](arg)

    def _next_step(self):
        if self.aborted:
            return

        self._cur_step += 1
        logging.debug('Entering step %d of %d in task %s.', self._cur_step + 1, len(self._steps), self.__class__.__name__)

        if self._cur_step >= len(self._steps):
            # That was the last one.
            self._sdone = True
            return

        # Call the init routine.
        self._done.set()
        self._steps[self._cur_step][0]()
        self._done.clear()

        with self._result_lock:
            self._results = []

        # Wake all free workers.
        self._master.wake_workers()



## Copyright 2017 Knossos authors, see NOTICE file
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##     http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.

from __future__ import absolute_import, print_function

import sys
import os.path
import logging
import re
import json
import stat
import sqlite3
import shutil
import semantic_version

from threading import Thread
from datetime import datetime
from .qt import QtCore, QtGui, QtWidgets, QtWebChannel
from . import center, runner, repo, windows, tasks, util, settings, nebula, clibs

if not QtWebChannel:
    from .qt import QtWebKit


class WebBridge(QtCore.QObject):
    _path = None

    asyncCbFinished = QtCore.Signal(int, str)
    showWelcome = QtCore.Signal()
    showDetailsPage = QtCore.Signal('QVariant')
    showRetailPrompt = QtCore.Signal()
    showLaunchPopup = QtCore.Signal(str)
    showModDetails = QtCore.Signal(str)
    updateModlist = QtCore.Signal(str, str)
    modProgress = QtCore.Signal(str, float, str)
    settingsArrived = QtCore.Signal(str)
    retailInstalled = QtCore.Signal()
    hidePopup = QtCore.Signal()
    applyDevDesc = QtCore.Signal(str)

    taskStarted = QtCore.Signal(float, str, list)
    taskProgress = QtCore.Signal(float, float, str)
    taskFinished = QtCore.Signal(float)
    taskMessage = QtCore.Signal(str)

    def __init__(self, webView=None):
        super(WebBridge, self).__init__()

        if QtWebChannel:
            self.bridge = self
            page = webView.page()
            channel = QtWebChannel.QWebChannel(page)

            page.setWebChannel(channel)
            channel.registerObject('fs2mod', self)

            if center.DEBUG and os.path.isdir('../html'):
                link = os.path.abspath('../html/index_debug.html')
                if sys.platform == 'win32':
                    link = '/' + link.replace('\\', '/')

                link = 'file://' + link
            else:
                link = 'qrc:///html/index.html'

            webView.load(QtCore.QUrl(link))

    @QtCore.Slot('QVariantList', result=str)
    def finishInit(self, tr_keys):
        trs = {}
        for k in tr_keys:
            trs[k] = QtCore.QCoreApplication.translate('modlist_ts', k)

        center.main_win.finish_init()
        return json.dumps({
            't': trs,
            'platform': sys.platform,
            'welcome': 'KN_WELCOME' in os.environ or center.settings['base_path'] is None
        })

    @QtCore.Slot(result=str)
    def getVersion(self):
        return center.VERSION

    @QtCore.Slot(result='QVariantList')
    def getMods(self):
        return list(center.mods.get())

    @QtCore.Slot(result='QVariantList')
    def getInstalledMods(self):
        return list(center.installed.get())

    @QtCore.Slot(result='QVariantMap')
    def getUpdates(self):
        updates = center.installed.get_updates()
        result = {}
        for mid, items in updates.items():
            versions = result[mid] = {}
            for ver_a, ver_b in items.items():
                versions[str(ver_a)] = str(ver_b)

        return result

    @QtCore.Slot(str, str, result=bool)
    def isInstalled(self, mid, spec=None):
        if spec is None:
            return mid in center.installed.mods
        else:
            spec = util.Spec(spec)
            mod = center.installed.mods.get(mid, None)
            if mod is None:
                return False

            return spec.match(mod.version)

    @QtCore.Slot(str, str, result='QVariantMap')
    def query(self, mid, spec=None):
        if spec is not None:
            if spec == '':
                spec = None
            else:
                if re.search(r'^\d+', spec):
                    spec = '==' + spec

                try:
                    spec = util.Spec(spec)
                except Exception:
                    logging.exception('Invalid spec "%s" passed to query()!', spec)
                    return -2

        try:
            return center.mods.query(mid, spec).get()
        except Exception:
            return None

    @QtCore.Slot()
    def fetchModlist(self):
        tasks.run_task(tasks.FetchTask())
        tasks.run_task(tasks.LoadLocalModsTask())

    @QtCore.Slot(bool, result='QVariantList')
    def requestModlist(self, async=False):
        if async:
            center.main_win.update_mod_list()
            return [None]
        else:
            return list(center.main_win.search_mods())

    @QtCore.Slot(str)
    def showTab(self, name):
        try:
            center.main_win.update_mod_buttons(name)
        except Exception:
            logging.exception('Failed to switch tabs!')

    @QtCore.Slot(str)
    def triggerSearch(self, term):
        center.main_win.perform_search(term)

    def _get_mod(self, mid, spec=None, mod_repo=None):
        if spec is not None:
            if spec == '':
                spec = None
            else:
                if re.search(r'^\d+', spec):
                    spec = '==' + spec

                try:
                    spec = util.Spec(spec)
                except Exception:
                    logging.exception('Invalid spec "%s" passed to a web API function!', spec)
                    return -2

        if mod_repo is None:
            mod_repo = center.installed

        try:
            return mod_repo.query(mid, spec)
        except repo.ModNotFound:
            logging.exception('Couldn\'t find mod "%s" (%s)!', mid, spec)
            return -1

    @QtCore.Slot(str, str, result=int)
    def showAvailableDetails(self, mid, spec=None):
        mod = self._get_mod(mid, spec, center.mods)
        if mod in (-1, -2):
            return mod

        self.showDetailsPage.emit(mod.get())
        return 0

    @QtCore.Slot(str, str, result=int)
    def showInstalledDetails(self, mid, spec=None):
        mod = self._get_mod(mid, spec)
        if mod in (-1, -2):
            return mod

        self.showDetailsPage.emit(mod.get())
        return 0

    @QtCore.Slot(str, str, 'QStringList', result=int)
    def install(self, mid, spec=None, pkgs=None):
        mod = self._get_mod(mid, spec, center.mods)
        if mod in (-1, -2):
            logging.debug('fs2mod.install(%s, %s) = %d', mid, spec, mod)
            return mod

        if pkgs is None:
            pkgs = []

        if mod.parent == 'FS2':
            retail = self._get_mod('FS2')

            if retail == -1:
                self.showRetailPrompt.emit()
                return 0

        windows.ModInstallWindow(mod, pkgs)
        return 0

    @QtCore.Slot(str, str, 'QStringList', result=int)
    def uninstall(self, mid, spec=None, pkgs=None):
        mod = self._get_mod(mid, spec)
        if mod in (-1, -2):
            return mod

        if mod.dev_mode:
            QtWidgets.QMessageBox.critical(None, 'Knossos', self.tr("I can't uninstall this mod because it's in dev mode!"))
            return 0

        if len(pkgs) == 0:
            plist = mod.packages
        else:
            plist = []
            pfound = set()
            for pkg in mod.packages:
                if pkg.name in pkgs:
                    plist.append(pkg)
                    pfound.add(pkg.name)

            if len(pfound) < len(pkgs):
                # Some packages are missing
                pmissing = set(pkgs) - pfound
                logging.warning('Missing packages %s.', ', '.join(pmissing))
                return -2

        titles = [pkg.name for pkg in plist if center.installed.is_installed(pkg)]
        # FIXME: Check if any other mod dependes on this mod before uninstalling it to avoid broken dependencies.

        deps = center.installed.get_dependents(plist)
        if deps:
            names = sorted([m.title for m in deps])
            msg = self.tr('You can\'t uninstall this because %s still depend on it.') % util.human_list(names)
            QtWidgets.QMessageBox.critical(None, 'Knossos', msg)
            return False

        msg = QtWidgets.QMessageBox()
        msg.setIcon(QtWidgets.QMessageBox.Question)
        msg.setText(self.tr('Do you really want to uninstall %s?') % (mod.title,))

        if len(titles) > 0:
            msg.setInformativeText(self.tr('%s will be removed.') % (', '.join(titles)))

        msg.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        msg.setDefaultButton(QtWidgets.QMessageBox.No)

        if msg.exec_() == QtWidgets.QMessageBox.Yes:
            tasks.run_task(tasks.UninstallTask(plist, mods=[mod]))
            return True
        else:
            return False

    @QtCore.Slot(str, str, result=int)
    def updateMod(self, mid, spec=None):
        mod = self._get_mod(mid, spec)

        if mod in (-1, -2):
            return mod

        all_vers = list(center.installed.query_all(mid))
        if len(all_vers) == 1 and not all_vers[0].dev_mode:
            # Only one version is installed, let's update it.
            tasks.run_task(tasks.UpdateTask(mod))
        else:
            # Just install the new version
            cur_pkgs = list(mod.packages)
            for i, pkg in enumerate(cur_pkgs):
                cur_pkgs[i] = center.mods.query(mod.mid, None, pkg.name)

            tasks.run_task(tasks.InstallTask(cur_pkgs, cur_pkgs[0].get_mod()))

        center.main_win.update_mod_buttons('progress')
        return 0

    @QtCore.Slot(float)
    def abortTask(self, tid):
        if hasattr(center.main_win, 'abort_task'):
            center.main_win.abort_task(int(tid))

    @QtCore.Slot(str, str, result=int)
    def runMod(self, mid, spec=None):
        mod = self._get_mod(mid, spec)
        if mod in (-1, -2):
            return mod

        runner.run_mod(mod)
        return 0

    @QtCore.Slot(str, str, result=list)
    def getModTools(self, mid, spec):
        mod = self._get_mod(mid, spec)
        if mod in (-1, -2):
            return [mod]

        labels = set()
        try:
            for exe in mod.get_executables():
                if exe.get('label') is not None:
                    labels.add(exe['label'])
        except repo.NoExecutablesFound:
            pass

        labels = list(labels)
        labels.sort()
        return labels

    @QtCore.Slot(str, str, str, str, str, result=int)
    def runModTool(self, mid, spec, tool, tool_spec, label):
        mod = self._get_mod(mid, spec)
        if mod in (-1, -2):
            return mod

        if tool == '':
            tool = None
        else:
            tool = self._get_mod(tool, tool_spec)
            if tool in (-1, -2):
                return mod

        if label == '':
            label = None

        runner.run_mod(mod, tool, label)
        return 0

    @QtCore.Slot(str, str, str, list)
    def runModAdvanced(self, mid, version, exe, mod_flag):
        mod = self._get_mod(mid, version)
        if mod in (-1, -2):
            return

        runner.run_mod_ex(mod, exe, mod_flag)

    @QtCore.Slot(str, str, result=int)
    def vercmp(self, a, b):
        try:
            a = semantic_version.Version(a)
            b = semantic_version.Version(b)
        except Exception:
            # logging.exception('Someone passed an invalid version to vercmp()!')
            return 0

        return a.__cmp__(b)

    @QtCore.Slot(str)
    def openExternal(self, link):
        if ':\\' in link:
            link = QtCore.QUrl.fromLocalFile(link)
        else:
            link = QtCore.QUrl(link)

        QtGui.QDesktopServices.openUrl(link)

    @QtCore.Slot(str, str, result=str)
    def browseFolder(self, title, path):
        return QtWidgets.QFileDialog.getExistingDirectory(None, title, path)

    @QtCore.Slot(str, str, str, result=list)
    def browseFiles(self, title, path, filter_):
        res = QtWidgets.QFileDialog.getOpenFileNames(None, title, path, filter_)
        if res:
            return res[0]
        else:
            return []

    @QtCore.Slot(str, result=bool)
    def setBasePath(self, path):
        if os.path.isfile(path):
            QtWidgets.QMessageBox.critical(None, 'Knossos', self.tr('The selected path is not a directory!'))
            return False
        elif not os.path.isdir(path):
            result = QtWidgets.QMessageBox.question(None, 'Knossos',
                self.tr('The selected path does not exist. Should I create the folder?'))

            if result == QtWidgets.QMessageBox.Yes:
                os.makedirs(path)
            else:
                return False
        else:
            vp_path = util.ipath(os.path.join(path, 'root_fs2.vp'))
            if os.path.isfile(vp_path):
                QtWidgets.QMessageBox.critical(None, 'Knossos', self.tr("Please don't use an existing FS2 directory. It won't work!"))
                return False

        center.settings['base_path'] = os.path.abspath(path)
        center.save_settings()
        tasks.run_task(tasks.LoadLocalModsTask())
        return True

    @QtCore.Slot()
    def getSettings(self):
        def cb(res):
            self.settingsArrived.emit(json.dumps(res))

        settings.get_settings(cb)

    @QtCore.Slot(str, str)
    def saveSetting(self, key, value):
        try:
            value = json.loads(value)
        except Exception:
            logging.exception('Failed to decode new value for setting "%s"! (%s)' % (key, value))
        else:
            settings.save_setting(key, value)

    @QtCore.Slot(str)
    def saveFsoSettings(self, data):
        try:
            data = json.loads(data)
        except Exception:
            logging.exception('Failed to decode new FSO settings! (%s)' % data)
        else:
            settings.save_fso_settings(data)

    @QtCore.Slot(result=str)
    def getDefaultFsoCaps(self):
        flags = None

        if center.settings['fs2_bin']:
            try:
                flags = settings.get_fso_flags(center.settings['fs2_bin'])
            except Exception:
                logging.exception('Failed to fetch FSO flags!')

        try:
            return json.dumps(flags)
        except Exception:
            logging.exception('Failed to encode FSO flags!')

    @QtCore.Slot(result=str)
    def searchRetailData(self):
        # Huge thanks go to jr2 for discovering everything implemented here to detect possible FS2 retail installs.
        # --ngld

        folders = [r'C:\GOG Games\Freespace2', r'C:\Games\Freespace2', r'C:\Games\Freespace 2']

        reg = QtCore.QSettings(r'HKEY_CURRENT_USER\Software\Valve\Steam', QtCore.QSettings.NativeFormat)
        reg.setFallbacksEnabled(False)

        steam_path = reg.value('SteamPath')

        if not steam_path:
            logging.info('No SteamPath detected!')
        else:
            steam_config = os.path.join(steam_path, 'config/config.vdf')
            if not os.path.isfile(steam_config):
                logging.warning('config.vdf is not where I expected it!')
            else:
                folders.append(os.path.join(steam_config, 'steamapps', 'common', 'Freespace 2'))

                with open(steam_config, 'r') as stream:
                    for m in re.finditer(r'"BaseInstallFolder_[0-9]+"\s+"([^"]+)"', stream.read()):
                        folders.append(os.path.join(m.group(1), 'Freespace 2'))

        for path in (r'HKEY_LOCAL_MACHINE\SOFTWARE\WOW6432Node\GOG.com\GOGFREESPACE2', r'HKEY_LOCAL_MACHINE\SOFTWARE\GOG.com\GOGFREESPACE2'):
            reg = QtCore.QSettings(path, QtCore.QSettings.NativeFormat)
            reg.setFallbacksEnabled(False)
            gog_path = reg.value('PATH')

            if gog_path:
                folders.append(gog_path)

        gog_db = os.path.expandvars(r'%ProgramData%\GOG.com\Galaxy\storage\index.db')
        if os.path.isfile(gog_db):
            try:
                db = sqlite3.connect(gog_db)
                c = db.cursor()
                c.execute('SELECT localpath FROM Products WHERE productId = 5')
                row = c.fetchone()
                if row:
                    folders.append(row[0])
            except Exception:
                logging.exception('Failed to read GOG Galaxy DB!')

        for path in folders:
            if util.is_fs2_retail_directory(path):
                return path

        return ''

    @QtCore.Slot(str, result=bool)
    def copyRetailData(self, path):
        if util.is_fs2_retail_directory(path):
            tasks.run_task(tasks.GOGCopyTask(path, os.path.join(center.settings['base_path'], 'FS2')))
            return True
        elif os.path.isfile(path) and path.endswith('.exe'):
            tasks.run_task(tasks.GOGExtractTask(path, os.path.join(center.settings['base_path'], 'FS2')))
            return True
        elif os.path.isfile(path):
            QtWidgets.QMessageBox.critical(None, 'Knossos', self.tr('The selected path is not a directory and not an installer!'))
            return False
        else:
            QtWidgets.QMessageBox.critical(None, 'Knossos', self.tr('The selected path does not contain the retail files!'))
            return False

    @QtCore.Slot(result=str)
    def getRunningTasks(self):
        tasks = center.main_win.get_tasks()
        res = {}

        for t, task in tasks.items():
            res[t] = {
                'title': task.title,
                'mods': task.mods
            }

        try:
            return json.dumps(res)
        except Exception:
            logging.exception('Failed to encoding running tasks!')
            return 'null'

    @QtCore.Slot(str, str, str, str, str, str, result=bool)
    def createMod(self, ini_path, name, mid, version, mtype, parent):
        if mtype in ('mod', 'ext'):
            if parent != 'FS2':
                parent = self._get_mod(parent)

                if parent == -1:
                    QtWidgets.QMessageBox.critical(None, 'Knossos', self.tr('The selected parent TC is not valid!'))
                    return False
                else:
                    parent = parent.mid
        else:
            parent = None

        mod = repo.InstalledMod({
            'title': name,
            'id': mid,
            'version': version,
            'type': mtype,
            'parent': parent,
            'dev_mode': True
        })
        mod.generate_folder()

        if os.path.isdir(mod.folder):
            QtWidgets.QMessageBox.critical(None, 'Knossos', self.tr('There already exists a mod with the chosen ID!'))
            return False

        exists = False
        try:
            neb = nebula.NebulaClient()
            neb.login()
            exists = not neb.check_mod_id(mid, name)
        except nebula.InvalidLoginException:
            QtWidgets.QMessageBox.warning(None, 'Knossos',
                self.tr("Knossos couldn't check if your mod ID is unique because it couldn't connect to the Nebula. " +
                    "Continue at your own risk if you're sure it is unique, otherwise please abort."))
        except Exception:
            logging.exception('Failed to contact the nebula!')
            QtWidgets.QMessageBox.warning(None, 'Knossos',
                self.tr("Knossos couldn't check if your mod ID is unique because it couldn't connect to the Nebula. " +
                    "Continue at your own risk if you're sure it is unique, otherwise please abort."))

        if exists:
            QtWidgets.QMessageBox.critical(None, 'Knossos',
                self.tr('Your chosen mod ID is already being used by someone else. Please choose a different one.'))
            return

        upper_folder = os.path.dirname(mod.folder)
        if not os.path.isdir(upper_folder):
            if mod.mtype in ('tool', 'engine') and upper_folder.endswith('bin'):
                try:
                    os.mkdir(upper_folder)
                except Exception:
                    logging.exception('Failed to create binary folder! (%s)' % upper_folder)
                    QtWidgets.QMessageBox.critical(None, 'Knossos', self.tr('I could not create the folder for binaries!'))
                    return False
            else:
                logging.error('%s did not exist during mod creation! (parent = %s)' % (mod.folder, mod.parent))
                QtWidgets.QMessageBox.critical(None, 'Knossos', self.tr('The chosen parent does not exist! Something went very wrong here!!'))
                return False

        if ini_path != '':
            # We need the ini mod for determining where to pull the VPs from
            ini_mod = repo.IniMod()
            ini_mod.load(ini_path)

            if len(ini_mod.get_primary_list()) > 0:
                QtWidgets.QMessageBox.critical(None, 'Knossos', self.tr(
                    'Ini mods with a primary list are currently not supported for importing!'))
                return False

            # This dict will convert a known secondary list entry to a mod.json style dependency
            dependency_mapping = {
                "mediavps_2014": {
                    "id": "MVPS",
                    "version": "3.7.2",
                },
                "mediavps_3612": {
                    "id": "MVPS",
                    "version": "3.6.12",
                },
                "mediavps": {
                    "id": "MVPS",
                    "version": "3.6.10",
                },
            }

            package_dependencies = []

            for dependency in ini_mod.get_secondary_list():
                dependency = dependency.lower()  # The mapping only works for lower case
                if dependency not in dependency_mapping:
                    # An unknown dependency is just skipped
                    QtWidgets.QMessageBox.warning(None, 'Knossos',
                                                  self.tr(
                                                      'The mod.ini dependency %s is not known to Knossos and could not be converted to a mod.json dependency.')
                                                  % (dependency))
                    continue

                package_dependencies.append(dependency_mapping[dependency])

            task = tasks.VpExtractionTask(mod, ini_mod)

            def finish_import():
                for vp_file in task.get_results():
                    base_filename = os.path.basename(vp_file).replace(".vp", "")

                    pkg = repo.InstalledPackage({
                        'name': base_filename,
                        'status': 'required',
                        'folder': base_filename,
                        'dependencies': package_dependencies,
                        'is_vp': True
                    })
                    mod.add_pkg(pkg)

                center.installed.add_mod(mod)
                mod.update_mod_flag()
                mod.save()

                center.main_win.update_mod_list()

            task.done.connect(finish_import)
            tasks.run_task(task)

            return True
        else:
            os.mkdir(mod.folder)
            mod.save()

            center.installed.add_mod(mod)
            center.main_win.update_mod_list()

            return True

    @QtCore.Slot(str, str, str, str, result=int)
    def addPackage(self, mid, version, pkg_name, pkg_folder):
        mod = self._get_mod(mid, version)
        if mod in (-1, -2):
            return mod

        if not mod.dev_mode:
            QtWidgets.QMessageBox.critical(None, 'Knossos',
                self.tr("You can't edit \"%s\" because it isn't in dev mode!") % mod.title)
            return -1

        pkg = mod.add_pkg(repo.Package({'name': pkg_name}))
        pkg.folder = pkg_folder

        pkg_path = os.path.join(mod.folder, pkg_folder)
        if not os.path.isdir(pkg_path):
            os.mkdir(pkg_path)

        mod.save()
        center.main_win.update_mod_list()

        return len(mod.packages) - 1

    @QtCore.Slot(str, str, int, result=bool)
    def deletePackage(self, mid, version, idx):
        mod = self._get_mod(mid, version)
        if mod in (-1, -2):
            return False

        if idx < 0 or idx >= len(mod.packages):
            logging.error('Invalid index passed to deletePackage()!')
            return False

        if not mod.dev_mode:
            QtWidgets.QMessageBox.critical(None, 'Knossos',
                self.tr("You can't edit \"%s\" because it isn't in dev mode!") % mod.title)
            return False

        # TODO: Delete the package folder?
        del mod.packages[idx]
        mod.save()
        center.main_win.update_mod_list()

        return True

    @QtCore.Slot(str, result=str)
    def selectImage(self, old_path):
        if old_path == '':
            old_dir = None
        else:
            old_dir = os.path.dirname(old_path)

        new_path, used_filter = QtWidgets.QFileDialog.getOpenFileName(None, self.tr('Please select an image'), old_dir,
                                                                      self.tr('Image (*.png *.jpg *.jpeg *.gif *.bmp)'))

        if new_path:
            return new_path
        elif os.path.isfile(old_path):
            return old_path
        else:
            return ''

    @QtCore.Slot(str, result=list)
    def addPkgExe(self, folder):
        if sys.platform == 'win32':
            filter_ = self.tr('Executables (*.exe)')
        else:
            filter_ = '*'

        res = QtWidgets.QFileDialog.getOpenFileNames(None, self.tr('Please select one or more executables'),
            folder, filter_)

        if not res:
            return []
        else:
            return [os.path.relpath(item, folder) for item in res[0]]

    @QtCore.Slot(str, result=list)
    def findPkgExes(self, folder):
        result = []

        for path, dirs, files in os.walk(folder):
            for fn in files:
                fn = os.path.join(path, fn)

                if fn.endswith('.exe'):
                    result.append(fn)
                elif '.so' not in fn and os.stat(fn).st_mode & stat.S_IXUSR == stat.S_IXUSR:
                    result.append(fn)

        return [os.path.relpath(item, folder) for item in result]

    def _store_mod_images(self, mod, img, imlist):
        if isinstance(img, list):
            for i, item in enumerate(img):
                img[i] = self._store_mod_images(mod, item, imlist)

            return img

        path = os.path.join(mod.folder, 'kn_images')
        if not os.path.isdir(path):
            os.mkdir(path)

        if os.path.abspath(img).startswith(path):
            imlist.add(os.path.basename(img))
            return img

        name, ext = os.path.splitext(img)
        dest = os.path.join(path, util.gen_hash(img)[1] + ext)

        logging.debug('Copying image from %s to %s.', img, dest)
        shutil.copyfile(img, dest)

        imlist.add(os.path.basename(dest))
        return dest

    def _clean_mod_images(self, mod, images):
        path = os.path.join(mod.folder, 'kn_images')
        if not os.path.isdir(path):
            return

        for item in os.listdir(path):
            if item not in images:
                logging.debug('Removing %s from %s because it is no longer needed.', item, mod)
                os.unlink(os.path.join(path, item))

    @QtCore.Slot(str)
    def saveModDetails(self, data):
        try:
            data = json.loads(data)
        except Exception:
            logging.exception('Failed to decode mod details!')
            QtWidgets.QMessageBox.critical(None, 'Error', self.tr('Internal data inconsistency. Please try again.'))
            return

        mod = self._get_mod(data['id'], data['version'])
        if mod == -1:
            logging.error('Failed find mod "%s" during save!' % data['id'])
            QtWidgets.QMessageBox.critical(None, 'Error', self.tr('Failed to find the mod! Weird...'))
            return

        if not mod.dev_mode:
            QtWidgets.QMessageBox.critical(None, 'Knossos',
                self.tr("You can't edit \"%s\" because it isn't in dev mode!") % mod.title)
            return

        if mod.mtype == 'engine':
            mod.stability = data['stability']

        mod.title = data['title']
        mod.description = data['description']
        imlist = set()
        for prop in ('logo', 'tile', 'banner', 'screenshots', 'attachments'):
            if data[prop]:
                setattr(mod, prop, self._store_mod_images(mod, data[prop], imlist))
            elif isinstance(data[prop], list):
                setattr(mod, prop, [])
            else:
                setattr(mod, prop, None)

        self._clean_mod_images(mod, imlist)

        mod.release_thread = data['release_thread']
        mod.videos = []
        for line in data['video_urls'].split('\n'):
            line = line.strip()
            if line != '':
                mod.videos.append(line)

        if data['first_release']:
            try:
                mod.first_release = datetime.strptime(data['first_release'], '%Y-%m-%d')
            except ValueError:
                QtWidgets.QMessageBox.critical(None, 'Knossos', self.tr('The entered first release date is invalid!'))
                return

        if data['last_update']:
            try:
                mod.last_update = datetime.strptime(data['last_update'], '%Y-%m-%d')
            except ValueError:
                QtWidgets.QMessageBox.critical(None, 'Knossos', self.tr('The entered last update date is invalid!'))
                return

        mod.save()
        center.main_win.update_mod_list()

    @QtCore.Slot(str, str, str, str)
    def savePackage(self, mid, version, pkg_name, data):
        try:
            data = json.loads(data)
        except Exception:
            logging.exception('Failed to decode mod details!')
            QtWidgets.QMessageBox.critical(None, 'Error', self.tr('Internal data inconsistency. Please try again.'))
            return

        mod = self._get_mod(mid, version)
        if mod == -1:
            logging.error('Failed find mod "%s" during save!' % mid)
            QtWidgets.QMessageBox.critical(None, 'Error', self.tr('Failed to find the mod! Weird...'))
            return

        if not mod.dev_mode:
            QtWidgets.QMessageBox.critical(None, 'Knossos',
                self.tr("You can't edit \"%s\" because it isn't in dev mode!") % mod.title)
            return

        pkg = None
        for item in mod.packages:
            if item.name == pkg_name:
                pkg = item
                break

        if not pkg:
            logging.error('Failed to find package "%s" for mod "%s"!' % (pkg_name, mid))
            QtWidgets.QMessageBox.critical(None, 'Error', self.tr('Failed to find the package! Weird...'))
            return

        pkg.notes = data['notes']
        pkg.status = data['status']
        pkg.dependencies = data['dependencies']

        if mod.mtype in ('engine', 'tool'):
            pkg.is_vp = False
            pkg.environment = data['environment']
            pkg.executables = data['executables']
        else:
            pkg.is_vp = data['is_vp']
            pkg.environment = None
            pkg.executables = []

        mod.update_mod_flag()
        mod.save()
        center.main_win.update_mod_list()

    @QtCore.Slot(str, str, str, str)
    def saveModFsoDetails(self, mid, version, build, cmdline):
        mod = self._get_mod(mid, version)
        if mod == -1:
            logging.error('Failed find mod "%s" during save!' % mid)
            QtWidgets.QMessageBox.critical(None, 'Error', self.tr('Failed to find the mod! Weird...'))
            return

        if not mod.dev_mode:
            QtWidgets.QMessageBox.critical(None, 'Knossos',
                self.tr("You can't edit \"%s\" because it isn't in dev mode!") % mod.title)
            return

        build = build.split('#')
        if len(build) != 2:
            logging.error('saveModFsoDetails(): build is not correctly formatted! (%s)' % build)
        else:
            if build[0] == 'custom':
                mod.custom_build = build[1]
            else:
                mod.custom_build = None

                try:
                    exes = mod.get_executables()
                except repo.NoExecutablesFound:
                    done = False
                    for pkg in mod.packages:
                        if pkg.status == 'required':
                            pkg.dependencies.append({
                                'id': build[0],
                                'version': '>=' + build[1]
                            })
                            done = True
                            break

                    if not done:
                        QtWidgets.QMessageBox.critical(None, 'Error',
                            self.tr('Failed to save the selected FSO build. Make sure that you have at least one required' +
                            ' package!'))
                else:
                    old_build = exes[0]['mod']
                    done = False

                    for pkg in mod.packages:
                        for dep in pkg.dependencies:
                            if dep['id'] == old_build.mid:
                                dep['id'] = build[0]
                                dep['version'] = '>=' + build[1]
                                done = True
                                break

                        if done:
                            break

                    if not done:
                        logging.error('Failed to update build dependency for "%s"! WHY?!?! (old_build = %s, new_build = %s)'
                            % (mod, old_build, build[0]))

        mod.cmdline = cmdline
        mod.save()

        center.main_win.update_mod_list()

    @QtCore.Slot(str, str, str, str)
    def saveUserFsoDetails(self, mid, version, build, cmdline):
        mod = self._get_mod(mid, version)
        if mod == -1:
            logging.error('Failed find mod "%s" during save!' % mid)
            QtWidgets.QMessageBox.critical(None, 'Error', self.tr('Failed to find the mod! Weird...'))
            return

        build = build.split('#')
        if len(build) != 2:
            logging.error('saveModFsoDetails(): build is not correctly formatted! (%s)' % build)
        else:
            if build[0] == 'custom':
                mod.user_custom_build = build[1]
                mod.user_exe = None
            else:
                mod.user_custom_build = None
                mod.user_exe = build

        mod.user_cmdline = cmdline
        mod.save_user()

    @QtCore.Slot(str, str, list)
    def saveModFlag(self, mid, version, mod_flag):
        mod = self._get_mod(mid, version)
        if mod in (-1, -2):
            return

        mod.mod_flag = mod_flag
        mod.save()

        center.main_win.update_mod_list()

    @QtCore.Slot(str, str)
    def startUpload(self, mid, version):
        mod = self._get_mod(mid, version)
        if mod in (-1, -2):
            return

        tasks.run_task(tasks.UploadTask(mod))

    @QtCore.Slot(str, str)
    def nebLogin(self, user, password):
        client = nebula.NebulaClient()
        try:
            result = client.login(user, password)
        except Exception:
            result = False
            logging.exception('Failed to login to Nebula!')

        if result:
            QtWidgets.QMessageBox.information(None, 'Knossos', 'Login successful!')

            # TODO: Figure out a better way for this!
            center.settings['neb_user'] = user
            center.settings['neb_password'] = password
            center.save_settings()
        else:
            QtWidgets.QMessageBox.critical(None, 'Knossos', 'Login failed.')

    @QtCore.Slot(str, str, str)
    def nebRegister(self, user, password, email):
        client = nebula.NebulaClient()

        try:
            result = client.register(user, password, email)
        except Exception:
            result = False
            logging.exception('Failed to register to the Nebula!')

        if result:
            QtWidgets.QMessageBox.information(None, 'Knossos',
                'Registered. Please check your e-mail inbox for your confirmation mail.')
        else:
            QtWidgets.QMessageBox.critical(None, 'Knossos', 'Registration failed. Please contact ngld.')

    @QtCore.Slot(str)
    def nebResetPassword(self, user):
        client = nebula.NebulaClient()

        try:
            result = client.reset_password(user)
        except Exception:
            result = False
            logging.exception('Failed to reset Nebula password!')

        if result:
            QtWidgets.QMessageBox.information(None, 'Knossos', 'You should now receive a mail with a reset link. ' +
                'Remember to check your spam folder!')
        else:
            QtWidgets.QMessageBox.critical(None, 'Knossos', 'Request failed. Please contect ngld.')

    @QtCore.Slot(str, str, str, result=bool)
    def nebReportMod(self, mid, version, message):
        mod = self._get_mod(mid, version)
        if mod in (-1, -2):
            return False

        client = nebula.NebulaClient()
        try:
            client.report_release(mod, message)
        except Exception:
            logging.exception('Failed to send mod report!')
            QtWidgets.QMessageBox.critical(None, 'Knossos', 'Request failed. Please contect ngld.')
            return False
        else:
            QtWidgets.QMessageBox.information(None, 'Knossos',
                'Thanks for your report. We will act on it as soon as possible.')
            return True

    @QtCore.Slot(str, str)
    def nebDeleteMod(self, mid, version):
        mod = self._get_mod(mid, version)
        if mod in (-1, -2):
            return

        fine = False
        client = nebula.NebulaClient()
        try:
            client.delete_release(mod)
            fine = True
        except nebula.AccessDeniedException:
            QtWidgets.QMessageBox.critical(None, 'Knossos', "You can't do that!")
        except nebula.RequestFailedException as exc:
            if exc.args[0] == 'not found':
                QtWidgets.QMessageBox.information(None, 'Knossos',
                    "This mod hasn't been uploaded and thus can't be removed from the nebula.")
                fine = True
            else:
                QtWidgets.QMessageBox.critical(None, 'Knossos', 'Request failed. Please contect ngld.')
        except Exception:
            logging.exception('Failed to send mod report!')
            QtWidgets.QMessageBox.critical(None, 'Knossos',
                'Request failed. You might have problems connecting to fsnebula.org.')
        else:
            QtWidgets.QMessageBox.critical(None, 'Knossos', 'The release was successfully deleted.')

        if fine:
            result = QtWidgets.QMessageBox.question(None, 'Knossos', 'Should the local files be deleted?')
            if result == QtWidgets.QMessageBox.Yes:
                tasks.run_task(tasks.UninstallTask(mod.packages))

    @QtCore.Slot(result=str)
    def selectCustomBuild(self):
        if sys.platform == 'win32':
            filter_ = '*.exe'
        else:
            filter_ = '*'

        res = QtWidgets.QFileDialog.getOpenFileNames(None, 'Please select your FSO build', None, filter_)
        if res and len(res[0]) > 0:
            return res[0][0]
        else:
            return ''

    @QtCore.Slot(str, str, result=str)
    def getFsoBuild(self, mid, version):
        mod = self._get_mod(mid, version)

        if mod.custom_build:
            return 'custom#' + mod.custom_build

        try:
            for item in mod.get_executables():
                if item.get('label') is None:
                    mod = item['mod']
                    return mod.mid + '#' + str(mod.version)
        except repo.NoExecutablesFound:
            return ''
        except Exception:
            logging.exception('Failed to fetch executables!')

        return ''

    @QtCore.Slot(str, str, result=str)
    def getUserBuild(self, mid, version):
        mod = self._get_mod(mid, version)

        if mod.user_custom_build:
            return 'custom#' + mod.user_custom_build

        try:
            for item in mod.get_executables(user=True):
                if item.get('label') is None:
                    mod = item['mod']
                    return mod.mid + '#' + str(mod.version)
        except repo.NoExecutablesFound:
            return ''
        except Exception:
            logging.exception('Failed to fetch executables!')

        return ''

    @QtCore.Slot(str, str, int)
    def getFsoCaps(self, mid, version, cb_id):
        mod = None
        if mid != 'custom':
            mod = self._get_mod(mid, version)
            if mod in (-1, -2):
                return

        def helper():
            flags = None
            if not mod:
                flags = settings.get_fso_flags(version)
            else:
                try:
                    for exe in mod.get_executables():
                        if not exe['label']:
                            flags = settings.get_fso_flags(exe['file'])
                except repo.NoExecutablesFound:
                    pass
                except Exception:
                    logging.exception('Failed to fetch FSO flags!')

            try:
                self.asyncCbFinished.emit(cb_id, json.dumps(flags))
            except Exception:
                logging.exception('Failed to encode FSO flags!')

        Thread(target=helper).start()

    @QtCore.Slot(str, str, str, str, result=bool)
    def createModVersion(self, mid, version, dest_ver, method):
        mod = self._get_mod(mid, version)

        if not isinstance(mod, repo.InstalledMod):
            logging.error('Mod %s (%s) should have gotten a new version but was not found!' % (mid, version))

            QtWidgets.QMessageBox.critical(None, self.tr('Error'),
                self.tr("Somehow I lost the mod you're talking about! I'm sorry, this is a bug."))
            return False

        if not mod.dev_mode:
            QtWidgets.QMessageBox.critical(None, 'Knossos',
                self.tr("You can't edit \"%s\" because it isn't in dev mode!") % mod.title)
            return False

        old_ver = semantic_version.Version(version, partial=True)
        try:
            dest_ver = semantic_version.Version(dest_ver, partial=True)
        except ValueError:
            QtWidgets.QMessageBox.critical(None, self.tr('Error'),
                self.tr("The specified version number is invalid!"))
            return False
        except Exception:
            logging.exception('Failed to parse new version (%s)!' % dest_ver)
            QtWidgets.QMessageBox.critical(None, self.tr('Error'),
                self.tr("Failed to parse the new version! This is a bug."))
            return False

        if old_ver >= dest_ver:
            # TODO: Is this check too restrictive?
            QtWidgets.QMessageBox.critical(None, self.tr('Error'),
                self.tr("The new version has to be higher than the old version!"))
            return False

        new_mod = mod.copy()
        new_mod.version = dest_ver
        new_mod.generate_folder()

        if os.path.isdir(new_mod.folder):
            QtWidgets.QMessageBox.critical(None, self.tr('Error'),
                self.tr("The destination folder (%s) already exists! I won't overwrite an existing folder!") % new_mod.folder)
            return False

        if method == 'copy':
            os.mkdir(new_mod.folder)

            tasks.run_task(tasks.CopyFolderTask(mod.folder, new_mod.folder))
        elif method == 'rename':
            try:
                util.safe_rename(mod.folder, new_mod.folder)
            except OSError:
                logging.exception('Failed to rename mod folder for new version!')
                QtWidgets.QMessageBox.critical(None, self.tr('Error'),
                    self.tr('Failed to rename folder "%s"! Make sure that no other pogram has locked it.') % mod.folder)
                return False

            try:
                center.installed.remove_mod(mod)
            except repo.ModNotFound:
                logging.exception('The old mod is missing after rename! Did the copy fail?')
        elif method == 'empty':
            os.mkdir(new_mod.folder)

            for pkg in new_mod.packages:
                os.mkdir(os.path.join(new_mod.folder, pkg.folder))

        new_mod.save()
        center.installed.add_mod(new_mod)
        center.main_win.update_mod_list()

        return True

    @QtCore.Slot(int, int, str)
    def testVoice(self, voice, volume, text):
        Thread(target=clibs.speak, args=(voice, volume, text)).start()

    @QtCore.Slot(str)
    def showDescEditor(self, text):
        windows.DescriptionEditorWindow(text)

    @QtCore.Slot(str, result=str)
    def parseIniMod(self, path):
        mod = repo.IniMod()
        mod.load(path)
        return json.dumps(mod.get())

    @QtCore.Slot(str, str)
    def verifyModIntegrity(self, mid, version):
        mod = self._get_mod(mid, version)
        if mod in (-1, -2):
            return

        tasks.run_task(tasks.CheckFilesTask(mod.packages))

    @QtCore.Slot()
    def openScreenshotFolder(self):
        path = os.path.join(settings.get_fso_profile_path(), 'screenshots')

        if os.path.isdir(path):
            QtGui.QDesktopServices.openUrl(QtCore.QUrl.fromLocalFile(path))
        else:
            QtWidgets.QMessageBox.critical(None, 'Knossos', "The screenshot folder doesn't exist. Try taking screenshots before clicking this button!")

    @QtCore.Slot(result=str)
    def getDefaultBasePath(self):
        if sys.platform == 'win32':
            return 'C:\\Games\\FreespaceOpen'
        elif sys.platform == 'linux':
            return os.path.expanduser('~/games/FreespaceOpen')
        elif sys.platform == 'macos':
            return os.path.expanduser('~/Documents/Games/FreespaceOpen')
        else:
            return ''


if QtWebChannel:
    BrowserCtrl = WebBridge
else:
    class BrowserCtrl(object):
        _view = None
        _nam = None
        bridge = None

        def __init__(self, webView):
            self._view = webView
            self.bridge = WebBridge()

            settings = webView.settings()
            settings.setAttribute(QtWebKit.QWebSettings.DeveloperExtrasEnabled, True)

            frame = webView.page().mainFrame()
            frame.javaScriptWindowObjectCleared.connect(self.insert_bridge)

            link = 'qrc:///html/index.html'
            webView.load(QtCore.QUrl(link))

        def insert_bridge(self):
            frame = self._view.page().mainFrame()

            del self.bridge
            self.bridge = WebBridge()
            frame.addToJavaScriptWindowObject('fs2mod', self.bridge)
